module Subset(issub) where

import RegExp
import RegExpData
import Parser
import General

sub :: String -> String -> Bool
sub a b = issub (parse a) (parse b)

issub :: RegExp -> RegExp -> Bool

-- universally true rules, regardless
issub (Union     a) b       = and (map (`issub` b) a)
issub (Omega      ) b       = True

issub (Terminal  a) (Omega      ) = False
issub (Terminal  a) (Terminal  b) = (a == b)
issub (Terminal  a) (Union     b) = or (map (issub (Terminal a)) b)
issub (Terminal  a) (Kleene    b) = issub (Terminal a) b
issub (Terminal  a) (Concat    b) = case notEwp of
                                        [ ] -> or (map (issub (Terminal a)) b)
                                        [x] -> issub (Terminal a) x
                                        _   -> False
                                    where notEwp = filter (not . ewp) b



-- more dubious ad hoc rules
issub (Kleene    a) (Union     b) = or (map (issub (Kleene a)) b)
issub (Kleene    a) (Concat    b) = case notEwp of
                                        [ ] -> or (map (issub (Kleene a)) b)
                                        [x] -> issub (Kleene a) x
                                        _   -> False
                                    where notEwp = filter (not . ewp) b


issub (Kleene    a) (Terminal  b) = False
issub (Union     a) (Terminal  b) = and (map (`issub` (Terminal b)) a)

issub (Union     a) (Kleene    b) = and (map (`issub` (Kleene   b)) a)

issub (Kleene    a) (Kleene    b) = issub a (Kleene b)

issub (Kleene    a) (Omega      ) = False


issub (Concat   []) b             = issub (Kleene Omega) b
issub (Concat  [a]) b             = issub a b



-- partially true, only when no other rules are available
issub (Concat a) (Union b) = or (map (issub (Concat a)) b)
issub (Concat (a:as)) (Kleene b) | issub a (Kleene b) && issub (Concat as) (Kleene b) = True
issub (Concat a     ) (Kleene b) | issub (Concat a) b = True
issub (Concat (a:as)) (Concat (b:bs)) | a == b = issub (Concat as) (Concat bs)
issub (Concat a) (Concat b) | a == b = True

{-
issub (Concat    a) b             = if empty a then issub Omega b
                                    else if not (null dropped) then issub (Concat useful) b
                                    else newsub (Concat a) b
    where (useful, dropped) = partition blank a
    
    --concat now is not equal to 0, and has no excess ^ terms
    newsub (Concat a) (Terminal  b) = False
    newsub (Concat a) (Omega      ) = False
    newsub (Concat a) (Intersect b) = or (map (issub (Concat a)) b)
    
    
    
    where
        reduce :: RegExp -> (Bool, 
        


-}



issub a b = False --error (show a ++ " " ++ show b)

{-

issub a

subset a (Kleene b) = ewp a && issub a b
    where
        --is a a subset of b*
    issub :: RegExp -> RegExp -> Bool

    --easy mappings
    issub (Omega      ) b = True --this is only true because a already possesses ewp
    issub (Kleene    a) b = issub a b
    issub (Concat  [a]) b = issub a b
    issub (Intersect a) b = and (map (`issub` b) a)

    issub a    (Kleene b) = issub a b

    issub (Terminal a) (Omega      ) = False
    issub (Terminal a) (Terminal  b) = a == b
    issub (Terminal a) (Intersect b) = or (map (Terminal a `issub`) b)
    issub (Terminal a) (Concat    b) =
        if null res then
            or (map (Terminal a `issub`) b)
        else if justOne res then
            Terminal a `issub` head res
        else
            False
        where res = filter (not . ewp) b

    --now it gets very hard, try to split it up any way possible
    issub (Concat a) b = or (map attempt (real_splits a))
        where
            attempt (a1, a2) = (Concat a1 `issub` b) && (Concat a2 `issub` b)

    issub a b = error ("While in subset: " ++ show a ++ " (= " ++ show b)
    

    real_splits = drop_last . tail . all_splits

    drop_last []     = []
    drop_last [x]    = []
    drop_last (x:xs) = x : drop_last xs

    all_splits []     = [([], [])]
    all_splits (x:xs) = ([], x:xs) : map (\(a,b) -> (x:a,b)) (all_splits xs)

-}
