
module General(
    sort_unique, merge_unique,
    diff, sort_diff,
    split,
    partition,
    setEq,
    justOne,
    first,
    elemBy,
    ifNull,
    thr1, thr2, thr3,
    bundle,
) where

import List(sortBy)

sort_unique :: Ord a => [a] -> [a]
sort_unique []  = []
sort_unique [x] = [x]
sort_unique xs = merge_unique (sort_unique a) (sort_unique b)
    where (a,b) = split xs

split :: [a] -> ([a], [a])
split []  = ([], [])
split [a] = ([a], [])
split (a:b:xs) = (a:as, b:bs)
    where (as,bs) = split xs

-- both list a and b must be individually unique
merge_unique :: Ord a => [a] -> [a] -> [a]
merge_unique a  [] = a
merge_unique [] b  = b
merge_unique (a:as) (b:bs) =
    case compare a b of
        EQ -> a: merge_unique as bs
        LT -> a: merge_unique as (b:bs)
        GT -> b: merge_unique (a:as) bs


-- set difference, all the elements in the first set that are not in the second set
diff :: Eq a => [a] -> [a] -> [a]
diff []     _  = []
diff (x:xs) ys = if x `elem` ys then res else x : res
    where res = diff xs ys

-- set difference, all elements in the first set not in the second set
-- assume both sets are ordered and unique
sort_diff :: Ord a => [a] -> [a] -> [a]
sort_diff (a:as) (b:bs) =
    case compare a b of
        EQ ->     sort_diff as     bs
        LT -> a : sort_diff as     (b:bs)
        GT ->     sort_diff (a:as) bs
-- either as or bs is null
sort_diff as     _      = as


-- split a list into two lists, on some criteria, (False, True)
partition :: (a -> Bool) -> [a] -> ([a], [a])
partition _ [] = ([], [])
partition f (x:xs) = if f x then (a,x:b) else (x:a,b) where (a,b) = partition f xs


setEq :: Eq a => [a] -> [a] -> Bool
setEq [] []     = True
setEq [] _      = False
setEq (x:xs) ys = x `elem` ys && setEq xs (remove_x ys)
    where remove_x (a:as) = if a == x then as else a:remove_x as


justOne :: [a] -> Bool
justOne [x] = True
justOne  _  = False

first :: [a] -> Maybe a
first []     = Nothing
first (x:xs) = Just x

elemBy :: (a -> Bool) -> [a] -> Bool
elemBy f = foldl (\ x y -> x || f y) False

ifNull :: [a] -> a -> [a]
ifNull a b = if null a then [b] else a

-- functions to extract just one element
thr1 (x,_,_) = x
thr2 (_,x,_) = x
thr3 (_,_,x) = x

-- functions to extract any two as a new tuple
thr12 (x,y,_) = (x,y)
thr13 (x,_,y) = (x,y)
thr23 (_,x,y) = (x,y)


-- bundle up elements (side effect of sorting)
bundle :: (a -> a -> Ordering) -> [a] -> [[a]]
bundle f xs = bundle_up (sortBy f xs)
    where
        bundle_up []     = []
        bundle_up (x:xs) = bundle_up_arg [x] xs
            where bundle_up_arg ys [] = [ys]
                  bundle_up_arg ys@(y:_) (x:xs) = if f x y == EQ then bundle_up_arg (x:ys) xs else ys : bundle_up_arg [x] xs
